 /**
 * Title: Advanced12: practising with Pipes and RandomAccessFiles
 * Description:
 * Copyright:    Copyright (c) 2004
 * Company:      Morpheus
 * @author Stefano Adriani
 * @version 1.0
 */

package adriani.tutorial.advanced ;

import java.io.*;
import java.util.*;

/** This class is created twice (within the main() execution):<P>
 *
 *	1.	The <CODE>Listener</CODE> instance waits for the user's input (row and charNumber) and
 *		then send these paramters by calling browser.setState(row, index).<BR>
 *	2.	The <CODE>Browser</CODE> instance waits for these parameters (row and charNumber) and
 *		then open a TXT file, reads the data and write them into the pipe.<P>
 *
 *  Finally, the "Bitwise methods" inherited from Advanced11 are used to produce output.<BR>
 *  REMARK: a dummy protocol is necessary to specify how many data are written in the pipe !
 */
class Advanced12 extends Advanced11 implements Runnable {

	// CLASS ATTRIBUTES
	int word = 10 ;					// Length of the word to read
	int rowNumber ;					// The row number to be read by the Browser
	int charNumber ;				// The char number to be read by the Browser
	Thread theThread ;				// the Thread of this runnable object
	boolean running = true ;		// The endless loop control variable

	PipedInputStream input ;		// Here the Listener reads the word
	PipedOutputStream output ;		// Here the Browser writes the word
	Advanced12 browser ;			// A reference to the other instance of this class

	/**	The Main method: creates the Pipes and starts the Threads.			 */
	public static void main(String args[])    {

		// Creating and linking Pipes
		PipedInputStream 	input 	= new PipedInputStream() ;
		PipedOutputStream 	output 	= new PipedOutputStream() ;
		try 				 { output.connect(input) ;							}
		catch(IOException e) { System.err.println("Pipes not connected!") ; 	}

		// Creating the Browser instance
        Advanced12 browser = new Advanced12(output);
		Thread thread1 = new Thread(browser) ;
		browser.theThread = thread1  ;
		thread1.start() ;

		// Creating the Listener instance
        Advanced12 listener = new Advanced12(input, browser);
		Thread thread2 = new Thread(listener) ;
		listener.theThread = thread2  ;
		thread2.start() ;
    }

	/** Constructor for the Browser instance. */
    public Advanced12(PipedOutputStream pipe) {

		this.output = pipe ;
		this.rowNumber = -1 ;
		this.charNumber = -1 ;
	}

	/** Constructor for the Listener instance. 					*/
    public Advanced12(PipedInputStream pipe, Advanced12 other)	{

		this.input = pipe ;
		this.browser = other ;
	}

	/** Called when this thread is started and "handling" IOExceptions. */
    public void run() {

		try {
			if (output != null) browse() ;
				else if ((input != null) && (browser != null) ) listen() ;
		}
		catch(IOException e) { System.err.println("IO error") ; }
	}

	/**	Run only in the Listener instance: it reads data from the pipe. */
    public void listen() throws IOException {

		while (running) {

			// First: getting input values ( see Advanced11 for getNumber() )
			System.out.print("Enter row number: ") ;
			int row = getNumber() ;
			System.out.print("Enter char number: ") ;
			int index = getNumber() ;
			if (row == 0) break ;				// This thread exits when we type "0"
			browser.setState(row, index) ;		// Sending request to the Browser

			// Second: reading values from the pipe
			String line = "WORD=" ;
			int bytes = input.read() ;						// Dummy protocol: 1st data is "length"
			System.out.println("Length: " + bytes) ;
			for (int j = 0 ; j < bytes; j++) {
				int value = input.read() ;
				if (value != 10) line += (char) value ;		// Avoiding the NewLine char
					else line += " " ;
				// Producing binary output (and formatting)
				if (j < 10) System.out.print(" ") ;
				System.out.print( j + ": " + stringValue(convert(value))) ;		// See Advanced11
				System.out.println(" (" + value + ")") ;
			}
			System.out.println(line + ".\n") ;				// Showing "end of word" with a dot
		}

		browser.running = false ;		// Ending the other thread
		input.close() ;					// Closing the PipedInputStream
	}

	/**	Run only in the Browser instance: it writes data in the pipe. */
    public void browse() throws IOException {

		// Opening the file
		String fp = File.separator ;
		String path = "adriani" + fp + "tutorial" + fp + "advanced" ;
		String name = path + fp + "Advanced12.java";
		RandomAccessFile file = new RandomAccessFile(name, "r") ;		// Read only access
		long length = file.length() ;

		String line = null ;
		String noLine = "NO LINE (end of file)" ;
		String emptyLine = "EMPTY LINE (no data)" ;
		while (running) {

			// The Browser works only "on request" ( see setState() )
			if (rowNumber > 0) {

				// First: Founding the row (1st line has index 0)
				rowNumber-- ;
				long cursor = - 1 ;
				for (int i = 0 ; i <= rowNumber; i++) {
					cursor = file.getFilePointer() ;
					line = file.readLine() ;
				}
				int size = -1 ;
				if (line != null ) size = line.length() ;

				// Second: Founding the word
				if (line != null) {
					if (size > 0) {
						file.seek(cursor + Math.min(charNumber, size)) ;
						int bytes = Math.min(word, size) ;
						output.write(bytes) ;				// Dummy protocol: 1st data is "length"
						for (int j = 0 ; j < bytes; j++) {
							byte b = file.readByte() ;
							output.write(b) ;
						}
					} else if (size == 0) {
						output.write(word) ;				// Dummy protocol: 1st data is "length"
						output.write(emptyLine.getBytes(), 0, word) ;
					}
				} else {
					output.write(word) ;
					output.write(noLine.getBytes(), 0, word) ;
				}

				// Third: restoring initial situation
				rowNumber = -1 ;
				file.seek(0) ;
			}
		}

		file.close() ;			// Closing the RandomAccessFile
		output.close() ;		// Closing the PipedOutputStream
	}

	/**	Used by the Browser to receive parameters. */
    public void setState(int row, int index)    	{

		this.rowNumber = row ;
		this.charNumber = index ;
	}

}	// End of class